#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .metaclasses import _GratingBase
from .simulation import *


class Grating3D(_GratingBase, Simulation):
    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        source=None,
        boundary_conditions=None,
        polarization=None,
        degree=1,
        pml_stretch=1 - 1j,
        periodic_map_tol=1e-8,
        eps_bc=1e-8,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Layered3D)
        assert source.dim == 3

        self.period = geometry.period
        self.epsilon = epsilon
        self.mu = mu
        self.degree = degree
        self.periodic_map_tol = periodic_map_tol
        self.eps_bc = eps_bc
        self.periodic_bcs = BiPeriodicBoundary3D(
            self.period,
            map_tol=self.periodic_map_tol,
            eps=self.eps_bc,
        )
        function_space = ComplexFunctionSpace(
            geometry.mesh, "N1curl", degree, constrained_domain=self.periodic_bcs
        )
        pml_bottom = PML(
            "z",
            stretch=pml_stretch,
            matched_domain="substrate",
            applied_domain="pml_bottom",
        )
        pml_top = PML(
            "z",
            stretch=pml_stretch,
            matched_domain="superstrate",
            applied_domain="pml_top",
        )

        epsilon_coeff = Coefficient(
            epsilon,
            geometry,
            pmls=[pml_bottom, pml_top],
            degree=degree,
            dim=3,
        )
        mu_coeff = Coefficient(
            mu, geometry, pmls=[pml_bottom, pml_top], degree=degree, dim=3
        )

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["substrate", "superstrate", "pml_bottom", "pml_top"]
        source_domains = [
            dom for dom in geometry.domains if dom not in no_source_domains
        ]

        formulation = Maxwell3DPeriodic(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="superstrate",
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

    def solve_system(self, again=False):
        uper = super().solve_system(again=again, vector_function=False)
        u_annex = self.formulation.annex_field["as_subdomain"]["stack"]
        u = uper * self.formulation.phasor
        self.solution = {"periodic": uper, "diffracted": u, "total": u + u_annex}
        return u

    def diffraction_efficiencies(
        self,
        N_d_order=0,
        cplx_effs=False,
        orders=False,
        subdomain_absorption=False,
        verbose=False,
    ):
        orders_num = np.linspace(
            -N_d_order,
            N_d_order,
            2 * N_d_order + 1,
        )

        k, gamma = {}, {}
        for d in ["substrate", "superstrate"]:
            k[d] = self.source.wavenumber * np.sqrt(
                complex(self.epsilon[d] * self.mu[d])
            )
            gamma[d] = np.conj(
                np.sqrt(
                    k[d] ** 2
                    - self.formulation.propagation_vector[0] ** 2
                    - self.formulation.propagation_vector[1] ** 2
                )
            )

        Phi = self.formulation.annex_field["phi"]
        r_annex = Phi[0][1::2]
        t_annex = Phi[-1][::2]
        eff_annex = dict(substrate=t_annex, superstrate=r_annex)
        Eper = self.solution["periodic"]
        effn = []
        effn_cplx = []
        for n in orders_num:
            effm = []
            effm_cplx = []
            for m in orders_num:
                if verbose:
                    print("*" * 55)
                    print(f"order ({n},{m})")
                    print("*" * 55)
                delta = 1 if n == m == 0 else 0
                qn = n * 2 * np.pi / self.period[0]
                pm = m * 2 * np.pi / self.period[1]
                alpha_n = self.formulation.propagation_vector[0] + qn
                beta_m = self.formulation.propagation_vector[1] + pm
                efficiencies = {}
                efficiencies_complex = {}
                for d in ["substrate", "superstrate"]:
                    s = 1 if d == "superstrate" else -1
                    # s = 1 if d == "substrate" else -1
                    gamma_nm = np.sqrt(k[d] ** 2 - alpha_n**2 - beta_m**2)
                    ph_x = phasor(
                        -qn, direction=0, degree=self.degree, domain=self.mesh
                    )
                    ph_y = phasor(
                        -pm, direction=1, degree=self.degree, domain=self.mesh
                    )
                    ph_z = phasor(
                        s * gamma_nm.real,
                        direction=2,
                        degree=self.degree,
                        domain=self.mesh,
                    )
                    ph_xy = ph_x * ph_y
                    Jnm = [
                        assemble(Eper[comp] * ph_xy * ph_z * self.dx(d))
                        / (self.period[0] * self.period[1])
                        for comp in range(3)
                    ]
                    ph_pos = np.exp(
                        -1 * 1j * gamma_nm * self.geometry.z_position["superstrate"]
                    )
                    ph_pos = 1 if d == "superstrate" else ph_pos
                    eff, sqnorm_eff = [], 0
                    for comp in range(3):
                        eff_ = (
                            delta * eff_annex[d][comp]
                            + Jnm[comp] / self.geometry.thicknesses[d]
                        ) * ph_pos
                        sqnorm_eff += eff_ * eff_.conj
                        eff.append(eff_)
                    eff_nrj = sqnorm_eff * gamma_nm / (gamma["superstrate"])
                    efficiencies_complex[d] = eff
                    efficiencies[d] = eff_nrj

                effm.append(efficiencies)
                effm_cplx.append(efficiencies_complex)
            effn.append(effm)
            effn_cplx.append(effm_cplx)

        Q, Qdomains = self.compute_absorption(subdomain_absorption=subdomain_absorption)

        T_nm = [[e["substrate"].real for e in b] for b in effn]
        R_nm = [[e["superstrate"].real for e in b] for b in effn]

        t_nm = [[e["substrate"] for e in b] for b in effn_cplx]
        r_nm = [[e["superstrate"] for e in b] for b in effn_cplx]

        T = sum(sum(_) for _ in T_nm)
        R = sum(sum(_) for _ in R_nm)

        B = R + T + Q

        effs = {
            "R": r_nm if cplx_effs else R_nm if orders else R,
            "T": t_nm if cplx_effs else T_nm if orders else T,
            "Q": Qdomains if subdomain_absorption else Q,
            "B": B,
        }
        if verbose:
            print("  Energy balance")
            print(f"  R = {R:0.6f}")
            print(f"  T = {T:0.6f}")
            print(f"  Q = {Q:0.6f}")
            print("  ------------------------")
            print(f"  B = {B:0.6f}")

        return effs

    def compute_absorption(self, subdomain_absorption=False):
        P0 = (
            self.period[0]
            * self.period[1]
            * (epsilon_0 / mu_0) ** 0.5
            * (np.cos(self.source.angle[0]))
            / 2
        )
        doms_no_pml = [
            z for z in self.epsilon.keys() if z not in ["pml_bottom", "pml_top"]
        ]
        Etot = self.solution["total"]
        # curl E = i ω μ_0 μ H
        inv_mu = self.formulation.mu.invert().as_subdomain()
        Htot = inv_mu / (1j * self.source.pulsation * mu_0) * curl(Etot)
        Qelec, Qmag = {}, {}
        if subdomain_absorption:
            for d in doms_no_pml:
                if np.all(self.epsilon[d].imag) == 0:
                    Qelec[d] = 0
                else:
                    # Etot = (
                    #     self.formulation.annex_field["as_dict"]["stack"][d]
                    #     + self.solution["diffracted"]
                    # )
                    elec_nrj_dens = dolfin.Constant(
                        0.5 * epsilon_0 * self.source.pulsation
                    ) * dot(self.epsilon[d] * Etot, Etot.conj)
                    Qelec[d] = -assemble(elec_nrj_dens * self.dx(d)).imag / P0
                if np.all(self.mu[d].imag) == 0:
                    Qmag[d] = 0
                else:
                    mag_nrj_dens = dolfin.Constant(
                        0.5 * mu_0 * self.source.pulsation
                    ) * dot(self.mu[d] * Htot, Htot.conj)
                    Qmag[d] = -assemble(mag_nrj_dens * self.dx(d)).imag / P0
            Q = sum(Qelec.values()) + sum(Qmag.values())
        else:
            epsilon_coeff = self.formulation.epsilon.as_subdomain()
            mu_coeff = self.formulation.mu.as_subdomain()
            elec_nrj_dens = dot(epsilon_coeff * Etot, Etot.conj)
            Qelec = (
                -0.5
                * epsilon_0
                * self.source.pulsation
                * assemble(elec_nrj_dens * self.dx(doms_no_pml))
                / P0
            ).imag
            mag_nrj_dens = dot(mu_coeff * Htot, Htot.conj)
            Qmag = (
                -0.5
                * mu_0
                * self.source.pulsation
                * assemble(mag_nrj_dens * self.dx(doms_no_pml))
                / P0
            ).imag
            Q = Qelec + Qmag
        Qdomains = {"electric": Qelec, "magnetic": Qmag}
        self.Qtot = Q
        self.Qdomains = Qdomains
        return Q, Qdomains



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import numpy as np

from gyptis.utils.helpers import *


def test_all():
    assert tanh(1.2) == np.tanh(1.2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


"""
Geometry definition using Gmsh api.
For more information see Gmsh's `documentation <https://gmsh.info/doc/texinfo/gmsh.html>`_
"""

import numbers
import os
import re
import sys
import tempfile
from collections import OrderedDict
from functools import wraps

import gmsh
import numpy as np

from .. import dolfin
from ..measure import Measure
from ..mesh import *
from ..plot import *

_geometry_module = sys.modules[__name__]
geo = gmsh.model.geo
occ = gmsh.model.occ
setnum = gmsh.option.setNumber
gmsh_options = gmsh.option


def _set_opt_gmsh(name, value):
    if isinstance(value, str):
        return gmsh_options.setString(name, value)
    elif isinstance(value, (numbers.Number, bool)):
        if isinstance(value, bool):
            value = int(value)
        return gmsh_options.setNumber(name, value)
    else:
        raise ValueError("value must be string or number")


def _get_opt_gmsh(name):
    try:
        return gmsh_options.getNumber(name)
    except Exception:
        return gmsh_options.getString(name)


setattr(gmsh_options, "set", _set_opt_gmsh)
setattr(gmsh_options, "get", _get_opt_gmsh)


def _add_method(cls, func, name):
    @wraps(func)
    def wrapper(*args, sync=True, **kwargs):
        out = func(*args, **kwargs)
        if sync:
            occ.synchronize()
        return out

    setattr(cls, name, wrapper)
    return func


def _dimtag(tag, dim=3):
    if not isinstance(tag, list):
        tag = [tag]
    return [(dim, t) for t in tag]


def _get_bnd(idf, dim):
    out = gmsh.model.getBoundary(_dimtag(idf, dim=dim), False, False, False)
    return [b[1] for b in out]


def _convert_name(name):
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


class Geometry:
    """Base class for geometry models."""

    def __init__(
        self,
        model_name="geometry",
        mesh_name="mesh.msh",
        data_dir=None,
        dim=3,
        gmsh_args=None,
        finalize=True,
        verbose=0,
        binary_mesh=True,
        options=None,
    ):
        if options is None:
            options = {}
        self.model_name = model_name
        self.mesh_name = mesh_name
        self.dim = dim
        self.subdomains = dict(volumes={}, surfaces={}, curves={}, points={})
        self.data_dir = data_dir or tempfile.mkdtemp()
        self.occ = occ
        self.mesh_object = {}
        self.measure = {}
        self.mesh = {}
        self.markers = {}
        self.options = options
        self.verbose = verbose
        self.binary_mesh = binary_mesh
        self.comm = dolfin.MPI.comm_world

        self.pml_physical = []

        for object_name in dir(occ):
            if (
                not object_name.startswith("__")
                and object_name != "mesh"
                and object_name not in dir(self)
            ):
                bound_method = getattr(occ, object_name)
                name = _convert_name(bound_method.__name__)
                _add_method(self, bound_method, name)

        self._gmsh_add_ellipse = self.add_ellipse
        del self.add_ellipse
        self._gmsh_add_circle = self.add_circle
        del self.add_circle
        self._gmsh_add_spline = self.add_spline
        del self.add_spline

        if finalize and gmsh.isInitialized():
            try:
                gmsh.finalize()
            except Exception:
                pass

        self.gmsh_args = gmsh_args
        if not gmsh.isInitialized():
            if gmsh_args is not None:
                gmsh.initialize(self.gmsh_args)
            else:
                gmsh.initialize()

        gmsh_options.set("General.Verbosity", self.verbose)
        gmsh_options.set("Mesh.Binary", self.binary_mesh)
        for k, v in options.items():
            gmsh_options.set(k, v)

    def _check_dim(self, dim):
        return self.dim if dim is None else dim

    def rotate(self, tag, point, axis, angle, dim=None):
        dt = self.dimtag(tag, dim=dim)
        return occ.rotate(dt, *point, *axis, angle)

    def add_physical(self, idf, name, dim=None):
        """Add a physical domain.

        Parameters
        ----------
        idf : int or list of int
            The identifiant(s) of elementary entities making the physical domain.
        name : str
            Name of the domain.
        dim : int
            Dimension.
        """
        dim = self._check_dim(dim)
        dicname = list(self.subdomains)[3 - dim]
        if not isinstance(idf, list):
            idf = [idf]
        num = gmsh.model.addPhysicalGroup(dim, idf)
        self.subdomains[dicname][name] = num
        gmsh.model.removePhysicalName(name)
        gmsh.model.setPhysicalName(dim, self.subdomains[dicname][name], name)
        return num

    def dimtag(self, idf, dim=None):
        """Convert an integer or list of integer to gmsh DimTag notation.

        Parameters
        ----------
        idf : int or list of int
            Label or list of labels.
        dim : type
            Dimension.

        Returns
        -------
        int or list of int
            A tuple (dim, tag) or list of such tuples (gmsh DimTag notation).

        """
        dim = self._check_dim(dim)
        return _dimtag(idf, dim=dim)

    def tagdim(self, x):
        if not isinstance(x, list):
            x = [x]
        return [t[1] for t in x]

    def _translation_matrix(self, t):
        M = [
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ]
        M[3], M[7], M[11] = t
        return M

    # def add_circle(self,x, y, z, ax, ay,**kwargs):
    #     ell = self._gmsh_add_ellipse(x, y, z, ax, ay,**kwargs)
    #     ell = self.add_curve_loop([ell])
    #     return self.add_plane_surface([ell])

    def add_circle(self, x, y, z, r, surface=True, **kwargs):
        if not surface:
            return self._gmsh_add_circle(x, y, z, r, **kwargs)
        circ = self._gmsh_add_circle(x, y, z, r, **kwargs)
        circ = self.add_curve_loop([circ])
        return self.add_plane_surface([circ])

    def add_ellipse(self, x, y, z, ax, ay, surface=True, **kwargs):
        if ax == ay:
            return self.add_circle(x, y, z, ax, surface=surface, **kwargs)
        elif ax < ay:
            ell = self.add_ellipse(x, y, z, ay, ax, surface=surface, **kwargs)
            self.rotate(ell, (x, y, z), (0, 0, 1), np.pi / 2, dim=2)
            return ell
        else:
            if not surface:
                return self._gmsh_add_ellipse(x, y, z, ax, ay, **kwargs)
            ell = self._gmsh_add_ellipse(x, y, z, ax, ay, **kwargs)
            ell = self.add_curve_loop([ell])
            return self.add_plane_surface([ell])

    def add_square(self, x, y, z, dx, **kwargs):
        return self.add_rectangle(x, y, z, dx, dx, **kwargs)

    def add_polygon(self, vertices, surface=True, mesh_size=0.0, **kwargs):
        """Adds a polygon.

        Parameters
        ----------
        points : array of shape (Npoints,3)
            Corrdinates of the points.
        mesh_size : float
            Mesh sizes at points (the default is 0.0).
        surface : type
            If True, creates a plane surface (the default is True).

        Returns
        -------
        int
            The tag of the polygon.

        """
        x, y, z = np.array(vertices).T
        N = len(vertices)
        points = []
        for i in range(N):
            p0 = self.add_point(x[i], y[i], z[i], meshSize=mesh_size)
            points.append(p0)
        lines = []
        for i in range(N - 1):
            lines.append(self.add_line(points[i], points[i + 1]))
        lines.append(self.add_line(points[i + 1], points[0]))
        loop = self.add_curve_loop(lines, **kwargs)
        return self.add_plane_surface([loop]) if surface else loop

    def add_spline(self, points, mesh_size=0.0, surface=True, **kwargs):
        """Adds a spline.

        Parameters
        ----------
        points : array of shape (Npoints,3)
            Corrdinates of the points.
        mesh_size : float
            Mesh sizes at points (the default is 0.0).
        surface : type
            If True, creates a plane surface (the default is True).

        Returns
        -------
        int
            The tag of the spline.

        """
        dt = [self.add_point(*p, meshSize=mesh_size) for p in points]
        if np.allclose(points[0], points[-1]):
            dt[-1] = dt[0]

        if not surface:
            return self._gmsh_add_spline(dt, **kwargs)
        spl = self._gmsh_add_spline(dt, **kwargs)
        spl = self.add_curve_loop([spl])
        return self.add_plane_surface([spl])

    def fragment(self, id1, id2, dim1=None, dim2=None, sync=True, map=False, **kwargs):
        dim1 = self._check_dim(dim1)
        dim2 = self._check_dim(dim2)
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        dimtags, mapping = occ.fragment(a1, a2, **kwargs)
        if sync:
            occ.synchronize()
        tags = [_[1] for _ in dimtags]
        return (tags, mapping) if map else tags

    def intersect(self, id1, id2, dim1=None, dim2=None, sync=True, map=False, **kwargs):
        dim1 = self._check_dim(dim1)
        dim2 = self._check_dim(dim2)
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        dimtags, mapping = occ.intersect(a1, a2, **kwargs)
        if sync:
            occ.synchronize()
        tags = [_[1] for _ in dimtags]
        return (tags, mapping) if map else tags

    def cut(self, id1, id2, dim1=None, dim2=None, sync=True, **kwargs):
        dim1 = self._check_dim(dim1)
        dim2 = self._check_dim(dim2)
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        ov, ovv = occ.cut(a1, a2, **kwargs)
        if sync:
            occ.synchronize()
        return [o[1] for o in ov]

    def fuse(self, id1, id2, dim1=None, dim2=None, sync=True):
        dim1 = self._check_dim(dim1)
        dim2 = self._check_dim(dim2)
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        ov, ovv = occ.fuse(a1, a2)
        if sync:
            occ.synchronize()
        return [o[1] for o in ov]

    def get_boundaries(self, idf, dim=None, physical=True):
        dim = self._check_dim(dim)
        if isinstance(idf, str):
            if dim == 2:
                type_entity = "surfaces"
            elif dim == 3:
                type_entity = "volumes"
            else:
                type_entity = "curves"
            idf = self.subdomains[type_entity][idf]

            n = gmsh.model.getEntitiesForPhysicalGroup(dim, idf)
            bnds = [_get_bnd(n_, dim=dim) for n_ in n]
            bnds = [item for sublist in bnds for item in sublist]
            return list(dict.fromkeys(bnds))
        else:
            n = gmsh.model.getEntitiesForPhysicalGroup(dim, idf)[0] if physical else idf
            return _get_bnd(n, dim=dim)

    def _set_size(self, idf, s, dim=None):
        dim = self._check_dim(dim)
        p = gmsh.model.getBoundary(
            self.dimtag(idf, dim=dim), False, False, True
        )  # Get all points
        gmsh.model.mesh.setSize(p, s)

    def _check_subdomains(self):
        groups = gmsh.model.getPhysicalGroups()
        names = [gmsh.model.getPhysicalName(*g) for g in groups]
        for subtype, subitems in self.subdomains.items():
            for idf in subitems.copy().keys():
                if idf not in names:
                    subitems.pop(idf)

    def set_mesh_size(self, params, dim=None):
        dim = self._check_dim(dim)
        if dim == 3:
            type_entity = "volumes"
        elif dim == 2:
            type_entity = "surfaces"
        elif dim == 1:
            type_entity = "curves"
        elif dim == 0:
            type_entity = "points"

        # revert sort so that smaller sizes are set last
        params = dict(
            sorted(params.items(), key=lambda item: float(item[1]), reverse=True)
        )

        for idf, p in params.items():
            if isinstance(idf, str):
                num = self.subdomains[type_entity][idf]
                n = gmsh.model.getEntitiesForPhysicalGroup(dim, num)
                for n_ in n:
                    self._set_size(n_, p, dim=dim)
            else:
                self._set_size(idf, p, dim=dim)

    def set_size(self, idf, s, dim=None):
        if hasattr(idf, "__len__") and not isinstance(idf, str):
            for i, id_ in enumerate(idf):
                s_ = s[i] if hasattr(s, "__len__") else s
                params = {id_: s_}
                self.set_mesh_size(params, dim=dim)
        else:
            self.set_mesh_size({idf: s}, dim=dim)

    def read_mesh_info(self):
        if self.dim == 1:
            marker_dim = "line"
            sub_dim = "curves"
            marker_dim_minus_1 = "point"
            sub_dim_dim_minus_1 = "points"
        elif self.dim == 2:
            marker_dim = "triangle"
            sub_dim = "surfaces"
            marker_dim_minus_1 = "line"
            sub_dim_dim_minus_1 = "curves"
        else:
            marker_dim = "tetra"
            sub_dim = "volumes"
            marker_dim_minus_1 = "triangle"
            sub_dim_dim_minus_1 = "surfaces"

        self.measure["dx"] = Measure(
            "dx",
            domain=self.mesh_object["mesh"],
            subdomain_data=self.mesh_object["markers"][marker_dim],
            subdomain_dict=self.subdomains[sub_dim],
        )

        # exterior_facets
        if (marker_dim_minus_1 in self.mesh_object["markers"].keys()) and (
            sub_dim_dim_minus_1 in self.subdomains.keys()
        ):
            self.measure["ds"] = Measure(
                "ds",
                domain=self.mesh_object["mesh"],
                subdomain_data=self.mesh_object["markers"][marker_dim_minus_1],
                subdomain_dict=self.subdomains[sub_dim_dim_minus_1],
            )

            # interior_facets

            self.measure["dS"] = Measure(
                "dS",
                domain=self.mesh_object["mesh"],
                subdomain_data=self.mesh_object["markers"][marker_dim_minus_1],
                subdomain_dict=self.subdomains[sub_dim_dim_minus_1],
            )
        else:
            self.measure["ds"] = None
            self.measure["dS"] = None

        self.mesh = self.mesh_object["mesh"]
        self.markers = self.mesh_object["markers"]

        if self.dim == 1:
            self.domains = self.subdomains["curves"]
            self.lines = {}
            self.markers = self.mesh_object["markers"]["line"]
            self.boundaries = {}
            self.boundary_markers = (
                self.mesh_object["markers"]["point"] if self.boundaries else []
            )

        elif self.dim == 2:
            self.domains = self.subdomains["surfaces"]
            self.lines = {}
            self.markers = self.mesh_object["markers"]["triangle"]
            self.boundaries = self.subdomains["curves"]
            self.boundary_markers = (
                self.mesh_object["markers"]["line"] if self.boundaries else []
            )

        else:
            self.domains = self.subdomains["volumes"]
            self.lines = self.subdomains["curves"]
            self.markers = self.mesh_object["markers"]["tetra"]
            self.boundaries = self.subdomains["surfaces"]
            self.boundary_markers = (
                self.mesh_object["markers"]["triangle"] if self.boundaries else []
            )

        self.points = self.subdomains["points"]
        self.unit_normal_vector = dolfin.FacetNormal(self.mesh)

    @property
    def msh_file(self):
        return os.path.join(self.data_dir, self.mesh_name)

    def _build_serial(
        self,
        interactive=False,
        generate_mesh=True,
        write_mesh=True,
        read_info=True,
        read_mesh=True,
        finalize=True,
        check_subdomains=True,
    ):
        if check_subdomains:
            self._check_subdomains()

        self.mesh_object = self.generate_mesh(
            generate=generate_mesh, write=write_mesh, read=read_mesh
        )

        if read_info:
            self.read_mesh_info()

        if interactive:
            gmsh.fltk.run()
        if finalize:
            gmsh.finalize()
        return self.mesh_object

    def build(
        self,
        interactive=False,
        generate_mesh=True,
        write_mesh=True,
        read_info=True,
        read_mesh=True,
        finalize=True,
        check_subdomains=True,
    ):
        """Build the geometry.

        Parameters
        ----------
        interactive : bool
            Open ``gmsh`` GUI? (the default is False).
        generate_mesh : type
            Mesh with ``gmsh``? (the default is True).
        write_mesh : type
            Write mesh to disk? (the default is True).
        read_info : type
            Read subdomain markers information? (the default is True).
        read_mesh : type
            Read mesh information? (the default is True).
        finalize : type
            Finalize ``gmsh`` API? (the default is True).
        check_subdomains : type
            Sanity check of subdomains names? (the default is True).

        Returns
        -------
        type
            A dictionary containing the mesh and markers.

        """
        if self.comm.size == 1:
            return self._build_serial(
                interactive=interactive,
                generate_mesh=generate_mesh,
                write_mesh=write_mesh,
                read_info=read_info,
                read_mesh=read_mesh,
                finalize=finalize,
                check_subdomains=check_subdomains,
            )
        if self.comm.rank == 0:
            self._build_serial(
                interactive=interactive,
                generate_mesh=generate_mesh,
                write_mesh=write_mesh,
                read_info=False,
                read_mesh=False,
                finalize=finalize,
                check_subdomains=check_subdomains,
            )
            tmp = self.data_dir
        else:
            tmp = None
        tmp = self.comm.bcast(tmp, root=0)
        self.data_dir = tmp
        self.mesh_object = self.read_mesh_file()
        self.read_mesh_info()

        return self.mesh_object

    def read_mesh_file(self, subdomains=None):
        if subdomains is not None:
            if isinstance(subdomains, str):
                subdomains = [subdomains]
            key = "volumes" if self.dim == 3 else "surfaces"
            subdomains_num = [self.subdomains[key][s] for s in subdomains]
        else:
            subdomains_num = subdomains

        return read_mesh(
            self.msh_file,
            data_dir=self.data_dir,
            dim=self.dim,
            subdomains=subdomains_num,
        )

    # def extract_sub_mesh(self, subdomains):
    #     return self.read_mesh_file(subdomains=subdomains)["mesh"]

    def extract_sub_mesh(self, subdomains):
        if self.comm.size == 1:
            key = "volumes" if self.dim == 3 else "surfaces"
            subdomains_num = self.subdomains[key][subdomains]
            return dolfin.SubMesh(self.mesh, self.markers, subdomains_num)
        outpath = (
            run_submesh(self, subdomains, outpath=None) if self.comm.rank == 0 else None
        )
        outpath = self.comm.bcast(outpath, root=0)
        return read_xdmf_mesh(outpath)

    def generate_mesh(self, generate=True, write=True, read=True):
        if generate:
            gmsh.model.mesh.generate(self.dim)
        if write:
            gmsh.write(self.msh_file)
        if read:
            return self.read_mesh_file()

    def plot_mesh(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        plt.sca(ax)
        return dolfin.plot(self.mesh, **kwargs)

    def plot_subdomains(self, markers=False, **kwargs):
        if markers:
            return plot_markers(self.markers, self.subdomains["surfaces"], **kwargs)
        else:
            return plot_subdomains(self.markers, **kwargs)

    def set_pml_mesh_size(self, s):
        for pml in self.pml_physical:
            self.set_mesh_size({pml: s})


def is_on_plane(P, A, B, C, eps=dolfin.DOLFIN_EPS):
    Ax, Ay, Az = A
    Bx, By, Bz = B
    Cx, Cy, Cz = C

    a = (By - Ay) * (Cz - Az) - (Cy - Ay) * (Bz - Az)
    b = (Bz - Az) * (Cx - Ax) - (Cz - Az) * (Bx - Ax)
    c = (Bx - Ax) * (Cy - Ay) - (Cx - Ax) * (By - Ay)
    d = -(a * Ax + b * Ay + c * Az)

    return dolfin.near(a * P[0] + b * P[1] + c * P[2] + d, 0, eps=eps)


def is_on_line(p, p1, p2, eps=dolfin.DOLFIN_EPS):
    x, y = p
    x1, y1 = p1
    x2, y2 = p2
    return dolfin.near((y - y1) * (x2 - x1), (y2 - y1) * (x - x1), eps=eps)


def is_on_line3D(p, p1, p2, eps=dolfin.DOLFIN_EPS):
    return is_on_plane(p, *p1, eps=eps) and is_on_plane(p, *p2, eps=eps)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import nlopt
import numpy as np

from . import dolfin as df
from .complex import *
from .materials import tensor_const
from .utils import array2function, function2array, mpi_print, project_iterative, tanh

df.parameters["allow_extrapolation"] = True


def simp(a, s_min=1, s_max=2, p=1, complex=True):
    """Solid isotropic material with penalisation (SIMP)"""
    if complex:
        return Complex(
            simp(a, s_min=s_min.real, s_max=s_max.real, p=p, complex=False),
            simp(a, s_min=s_min.imag, s_max=s_max.imag, p=p, complex=False),
        )
    else:
        return s_min + (s_max - s_min) * a**p


def projection(a, beta=1, nu=0.5):
    """Projection operator."""
    return (tanh(beta * nu) + tanh(beta * (a - nu))) / (
        tanh(beta * nu) + tanh(beta * (1 - nu))
    )


def projection_gradient(a, beta=1, nu=0.5):
    """Projection operator gradient dproj/da."""
    return (
        beta
        * (1 - (tanh(beta * (a - nu))) ** 2)
        / (tanh(beta * nu) + tanh(beta * (1 - nu)))
    )


class Filter:
    def __init__(
        self,
        rfilt=0,
        function_space=None,
        degree=1,
        solver=None,
        mesh=None,
        output_function_space=None,
    ):
        self.rfilt = rfilt
        self._rfilt_scaled = self.rfilt / (2 * 3**0.5)
        self.solver = solver
        self.degree = degree
        self._mesh = mesh
        self._function_space = function_space
        self.output_function_space = output_function_space

    def weak(self, a):
        self.mesh = a.function_space().mesh() if self._mesh is None else self._mesh
        self.dim = self.mesh.ufl_domain().geometric_dimension()
        self.function_space = self._function_space or df.FunctionSpace(
            self.mesh, "CG", self.degree
        )
        af = df.TrialFunction(self.function_space)
        vf = df.TestFunction(self.function_space)
        if hasattr(self.rfilt, "shape"):
            if np.shape(self.rfilt) in [(2, 2), (3, 3)]:
                self._rfilt_scaled = tensor_const(
                    self._rfilt_scaled, dim=self.dim, real=True
                )
            else:
                raise ValueError("Wrong shape for rfilt")
        else:
            self._rfilt_scaled = df.Constant(self._rfilt_scaled)

        lhs = (
            df.inner(
                self._rfilt_scaled * df.grad(af),
                self._rfilt_scaled * df.grad(vf),
            )
            * df.dx
            + df.inner(af, vf) * df.dx
        )
        rhs = df.inner(a, vf) * df.dx
        return lhs, rhs

    def apply(self, a):
        if np.all(self.rfilt == 0):
            return a
        lhs, rhs = self.weak(a)
        af = df.Function(self.function_space, name="Filtered density")
        self.vector = df.assemble(rhs)
        if self.solver is None:
            self.matrix = df.assemble(lhs)
            self.solver = df.KrylovSolver(self.matrix, "cg", "jacobi")
        self.solver.solve(af.vector(), self.vector)
        self.solution = af
        return (
            project_iterative(af, self.output_function_space)
            if self.output_function_space is not None
            else af
        )


def filtering(a, rfilt=0, function_space=None, degree=1, solver=None, mesh=None):
    return Filter(rfilt, function_space, degree, solver, mesh).apply(a)


def transfer_function(fromFunc, Vto):
    toFunc = df.Function(Vto)
    fromFunc.set_allow_extrapolation(True)
    A = df.PETScDMCollection.create_transfer_matrix(
        fromFunc.ufl_function_space(), toFunc.ufl_function_space()
    )
    # toFunc.vector()[:] = A * fromFunc.vector()

    A.mat().mult(fromFunc.vector().vec(), toFunc.vector().vec())

    return toFunc


def derivative(f, x, ctrl_space=None, array=False):
    dfdx = df.compute_gradient(f, df.Control(x))
    if ctrl_space is not None:
        dfdx = transfer_function(dfdx, ctrl_space)
    return function2array(dfdx) if array else dfdx


def transfer_sub_mesh(x, geometry, source_space, target_space, subdomain):
    markers = geometry.markers
    domains = geometry.domains
    a0 = df.Function(source_space)
    mdes = markers.where_equal(domains[subdomain])
    b = function2array(project_iterative(x, target_space))
    a = function2array(a0)
    comm = df.MPI.comm_world
    b = comm.gather(b, root=0)
    a = comm.gather(a, root=0)
    mdes = comm.gather(mdes, root=0)
    if df.MPI.rank(comm) == 0:
        mdes = np.hstack(mdes)
        b = np.hstack(b)
        a = np.hstack(a)
        mdes1 = [int(i) for i in mdes]
        a[mdes1] = b
        # sys.stdout.flush()
    else:
        a = None
    a = comm.bcast(a, root=0)
    return array2function(a, source_space)


class TopologyOptimizer:
    def __init__(
        self,
        fun,
        geometry,
        design="design",
        eps_bounds=(1, 3),
        p=1,
        rfilt=0,
        filtering_type="density",
        threshold=(0, 8),
        maxiter=20,
        stopval=None,
        ftol_rel=None,
        xtol_rel=None,
        callback=None,
        args=None,
        verbose=True,
    ):
        self.fun = fun
        self.design = design
        self.threshold = threshold
        self.rfilt = rfilt
        self.filtering_type = filtering_type
        self.maxiter = maxiter
        self.stopval = stopval
        self.ftol_rel = ftol_rel
        self.xtol_rel = xtol_rel
        self.callback = callback
        self.args = args or []
        self.verbose = verbose
        self.callback_output = []
        self.eps_min, self.eps_max = eps_bounds
        self.p = p
        self.geometry = geometry
        self.mesh = self.geometry.mesh
        self.submesh_plt = self.geometry.extract_sub_mesh(self.design)
        self.submesh = self.submesh_plt
        # self.submesh = self.mesh
        # self.submesh = df.SubMesh(self.mesh, self.geometry.markers, self.geometry.domains[self.design])

        self.fs_ctrl = df.FunctionSpace(self.mesh, "DG", 0)
        self.fs_sub = df.FunctionSpace(self.submesh, "DG", 0)
        self.nvar = self.fs_sub.dim()
        self.filter = Filter(self.rfilt, degree=1, output_function_space=self.fs_sub)

    def design_fun(self, density_fp):
        return simp(
            density_fp,
            Constant(self.eps_min),
            Constant(self.eps_max),
            df.Constant(self.p),
        )

    def _topopt_wrapper(
        self,
        objfun,
        Asub,
        Actrl,
        eps_design_min,
        eps_design_max,
        p=1,
        filter=None,
        filtering_type="density",
        proj_level=None,
        reset=True,
        grad=True,
        *args,
    ):
        def wrapper(x):
            proj = proj_level is not None
            filt = filter is not None
            if reset:
                df.set_working_tape(df.Tape())
            density = array2function(x, Asub)
            self.density = density
            filter.solver = None
            if filtering_type == "sensitivity":
                self.density_f = density
            else:
                self.density_f = filter.apply(density) if filt else density

            # density_f = transfer_sub_mesh(density_f, self.geometry, Actrl, Asub, self.design)

            ctrl = transfer_function(self.density_f, Actrl)
            self.density_fp = (
                projection(ctrl, beta=df.Constant(2**proj_level))
                if proj
                else self.density_f
            )
            self.epsilon_design = self.design_fun(self.density_fp)
            objective = objfun(self.epsilon_design, *args)

            self.objective = objective

            if grad:
                dobjective_dx = derivative(objective, ctrl)
                if filt and filtering_type == "sensitivity":
                    f = project_iterative(density * dobjective_dx, Asub)
                    dfdd = filter.apply(f)
                    dobjective_dx = dfdd / (density + 1e-3)
                dobjective_dx = project_iterative(dobjective_dx, Asub)
                dobjective_dx = function2array(dobjective_dx)
                self.dobjective_dx = dobjective_dx
                return objective, dobjective_dx
            else:
                return objective, None

        return wrapper

    def minimize(self, x0):
        self.x0 = x0
        if self.verbose:
            mpi_print("#################################################")
            mpi_print(f"Topology optimization with {self.nvar} variables")
            mpi_print("#################################################")
            mpi_print("")

        for iopt in range(*self.threshold):
            self.proj_level = iopt

            self.wrapper = self._topopt_wrapper(
                self.fun,
                self.fs_sub,
                self.fs_ctrl,
                self.eps_min,
                self.eps_max,
                self.p,
                self.filter,
                self.filtering_type,
                self.proj_level,
                reset=True,
                grad=True,
                *self.args,
            )

            self._cbout = []
            if self.verbose:
                mpi_print(f"global iteration {iopt}")
                mpi_print("---------------------------------------------")

            def _get_gradient(dy):
                comm = df.MPI.comm_world
                rank = df.MPI.rank(comm)
                dyall = comm.gather(dy, root=0)

                dy = np.hstack(dyall) if rank == 0 else np.empty(self.nvar)
                comm.Bcast(dy, root=0)

                return dy

            def fun_nlopt(x, gradn):
                y, dy = self.wrapper(x)
                if self.verbose:
                    mpi_print(f"objective = {y}")
                gradn[:] = _get_gradient(dy)
                if self.callback is not None:
                    out = self.callback(self)
                    self._cbout.append(out)
                return y

            lb = np.zeros(self.nvar, dtype=float)
            ub = np.ones(self.nvar, dtype=float)

            opt = nlopt.opt(nlopt.LD_MMA, self.nvar)
            opt.set_lower_bounds(lb)
            opt.set_upper_bounds(ub)
            if self.ftol_rel is not None:
                opt.set_ftol_rel(self.ftol_rel)
            if self.xtol_rel is not None:
                opt.set_xtol_rel(self.xtol_rel)
            if self.stopval is not None:
                opt.set_stopval(self.stopval)
            if self.maxiter is not None:
                opt.set_maxeval(self.maxiter)

            opt.set_min_objective(fun_nlopt)
            xopt = opt.optimize(x0)
            fopt = opt.last_optimum_value()
            if self.callback is not None:
                self.callback_output.append(self._cbout)
            x0 = xopt

        self.opt = opt
        self.xopt = xopt
        self.fopt = fopt
        return xopt, fopt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


import numpy as np

import gyptis as gy


def test_hchom2d():
    d = 1
    v = (d, 0), (0, d)
    a = d / 2
    lmin = a / 50

    lattice = gy.Lattice(dim=2, vectors=v)
    incl = lattice.add_square(a / 2, a / 2, 0, a)
    cell = lattice.cut(lattice.cell, incl)
    lattice.add_physical(cell, "background")
    lattice.set_size("background", lmin)
    lattice.build()

    inclusion = gy.Geometry(dim=2)
    incl = inclusion.add_square(a / 2, a / 2, 0, a)
    inclusion.add_physical(incl, "inclusion")
    bnds = inclusion.get_boundaries("inclusion")
    inclusion.add_physical(bnds, "inclusion_bnds", dim=1)
    inclusion.set_size("inclusion", lmin)
    inclusion.build()

    eps_i = 200 - 5j
    epsilon = dict(inclusion=eps_i, background=1)
    mu = dict(inclusion=1, background=1)

    def analytical_mueff(k, Nmax=10):
        nms = [2 * n + 1 for n in range(Nmax)]
        mu = 1
        for n in nms:
            for m in nms:
                knm = np.pi / a * (n**2 + m**2) ** 0.5
                qn = np.pi / a * n
                pm = np.pi / a * m
                alpha = 2 / qn * 2 / pm
                norm = (a / 2) ** 2
                mu += (
                    -(k**2 * eps_i) / (k**2 * eps_i - knm**2) * alpha**2 / norm
                )
        return mu

    hom = gy.models.HighContrastHomogenization2D(
        lattice, inclusion, epsilon, mu, degree=2
    )
    eps_eff = hom.get_effective_permittivity(scalar=True)
    print(eps_eff)
    assert np.abs(eps_eff[0][0] - 1.73) < 1e-3
    neigs = 50
    wavevector_target = 0.1
    lambdas = np.linspace(4, 15, 1000) * d
    k = 2 * np.pi / lambdas

    mu_eff = hom.get_effective_permeability(
        k, neigs=neigs, wavevector_target=wavevector_target
    ).tocomplex()
    mu_eff_ana = analytical_mueff(k)
    err = np.mean(np.abs(mu_eff - mu_eff_ana))
    print(err)
    assert np.allclose(mu_eff, mu_eff_ana, atol=5e-3)
    assert err < 1e-3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


import numpy as np
import pytest


def test_simple(tol=1e-12):
    from gyptis.complex import Complex, Constant

    x, y, x1, y1 = np.random.rand(4) - 0.5
    z = Complex(x, y)
    z1 = Complex(x1, y1)
    q = x + 1j * y
    q1 = x1 + 1j * y1
    assert str(z) == f"({x} + {y}j)"
    assert repr(z) == f"Complex({x}, {y})"
    assert z == q
    assert -z == Complex(-x, -y)
    assert z != Complex(x / 2, y)
    assert z != Complex(x, y / 2)
    assert z != Complex(x / 2, y / 2)
    assert z != Complex(x, y) / 2
    assert z != q / q1
    assert z != q / 2
    assert 2 * z == Complex(2 * x, 2 * y)
    assert z / z == 1
    assert 1 / z == 1 / z
    assert z * z1 == q * q1
    assert z * q1 == q * z1
    assert abs(z / q1 - q / z1) < tol
    assert abs(z / z1 - q / q1) < tol
    assert abs(z) == (x**2 + y**2) ** 0.5
    assert z.conj == q.conj()
    assert z.module == abs(z)
    assert z.phase == np.angle(q)
    assert abs(z**2 - q**2) < tol
    assert abs(z ** (-3.4) - q ** (-3.4)) < tol
    assert z + q1 == z1 + q
    assert len(Complex(1, 1)) == 0
    assert len(Complex([1, 2], [1, 2])) == 2
    assert Complex(np.random.rand(3, 2), np.random.rand(3, 2)).shape == (3, 2)

    with pytest.raises(ValueError):
        len(Complex([1], [3, 2]))

    with pytest.raises(NotImplementedError):
        z**z

    Z = Complex([1, 2], [3, 2])
    for i, r in enumerate(Z):
        assert r == Z[i]

    Constant((1 + 6 * 1j, 6 - 0.23j), name="test")
    Constant((1 + 6 * 1j, 6 - 0.23j))
    Constant(3 + 1j)
    Constant(12)
    Constant((1, 2))


def test_complex(tol=1e-15):
    import gyptis
    from gyptis import dolfin
    from gyptis.bc import DirichletBC

    nx, ny = 50, 50
    mesh = dolfin.UnitSquareMesh(nx, ny)

    def boundary(x):
        return (
            x[0] < dolfin.DOLFIN_EPS
            or x[0] > 1.0 - dolfin.DOLFIN_EPS
            or x[1] < dolfin.DOLFIN_EPS
            or x[1] > 1.0 - dolfin.DOLFIN_EPS
        )

    W = gyptis.ComplexFunctionSpace(mesh, "CG", 1)
    # W0 =  dolfin.FunctionSpace(mesh,W.ufl_element().extract_component(0)[1])
    W0 = dolfin.FunctionSpace(mesh, "DG", 0)

    # u = Function(W)
    utrial = gyptis.TrialFunction(W)
    utest = gyptis.TestFunction(W)
    dx = dolfin.dx(domain=mesh)

    k = 2
    qre, qim = 2, -1
    gamma = 0.05
    expr = f"exp(-pow((x[0]-0.5)/{gamma},2)-pow((x[1]-0.5)/{gamma},2))"
    expr_re = f"{qre}+{expr}"
    expr_im = f"{qim}+{expr}"
    sol = dolfin.Expression((expr_re, expr_im), degree=1, domain=mesh)
    sol = gyptis.Complex(sol[0], sol[1])

    source = (k**2) * sol * utest + gyptis.inner(gyptis.grad(sol), gyptis.grad(utest))

    F = (
        gyptis.inner(gyptis.grad(utrial), gyptis.grad(utest)) * dx
        + k**2 * utrial * utest * dx
        - source * dx
    )
    F = F.real + F.imag

    u_dirichlet = qre + 1j * qim
    bcre, bcim = DirichletBC(W, u_dirichlet, boundary)

    bcs = [bcre, bcim]
    Lh = dolfin.rhs(F)
    bh = gyptis.assemble(Lh)
    ah = dolfin.lhs(F)
    Ah = gyptis.assemble(ah)
    dolfin.VectorFunctionSpace(mesh, "CG", 2)

    # u = Function(W)

    u = dolfin.Function(W)

    def _solve(solver, u, tol, tol_local):
        for bc in bcs:
            bc.apply(Ah, bh)
        solver.solve(Ah, u.vector(), bh)

        u = gyptis.Complex(u[0], u[1])
        assert abs(gyptis.assemble((u - sol) * dx)) < tol
        assert abs((u((0.5, 0.5)) - sol((0.5, 0.5)))) < tol_local
        return u

    # _solve(dolfin.LUSolver(Ah), 1e-12, 1e-12)
    u = _solve(dolfin.PETScKrylovSolver(), u, 1e-6, 1e-4)

    gyptis.project(u.real, W0)
    gyptis.inner(u.real, u)
    gyptis.inner(u, u.imag)
    gyptis.inner(u.real, u.imag)

    u.phase
    u.module



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from math import pi

import pytest

from gyptis import BoxPML, dolfin
from gyptis.geometry import *


def geom2D(square_size=1, cyl_size=0.3, mesh_size=0.1):
    model = Geometry("Square", dim=2)
    box = model.add_rectangle(
        -square_size / 2, -square_size / 2, 0, square_size, square_size
    )
    cyl = model.add_rectangle(-cyl_size / 2, -cyl_size / 2, 0, cyl_size, cyl_size)
    cyl, box = model.fragment(cyl, box)

    model.add_physical(box, "box")
    model.add_physical(cyl, "cyl")

    outer_bnds = model.get_boundaries("box")[:-4]
    cyl_bnds = model.get_boundaries("cyl")
    model.add_physical(outer_bnds, "outer_bnds", dim=1)
    model.add_physical(cyl_bnds, "cyl_bnds", dim=1)
    model.set_size("box", 1 * mesh_size)
    model.set_size("cyl", 1 * mesh_size)
    model.set_size("cyl_bnds", mesh_size, dim=1)
    model.build()
    model.cyl_size = cyl_size
    model.square_size = square_size

    return model


def test_2D():
    model = geom2D(mesh_size=0.01)
    dx = model.measure["dx"]
    area_cyl = dolfin.assemble(1 * dx("cyl"))
    assert abs(area_cyl - model.cyl_size**2) < 1e-4


def test_3D():
    model = Geometry("Sphere")
    box = model.add_box(-1, -1, -1, 2, 2, 2)
    sphere = model.add_sphere(0, 0, 0, 0.5)
    sphere, box = model.fragment(sphere, box)
    model.set_size(box, 0.3)
    model.set_size(sphere, 0.3)
    model.add_physical(sphere, "sphere")
    model.add_physical(box, "box")
    outer_bnds = model.get_boundaries("box")
    model.add_physical(outer_bnds, "outer_bnds", dim=2)
    model.build()
    assert model.subdomains == {
        "volumes": {"sphere": 1, "box": 2},
        "surfaces": {"outer_bnds": 3},
        "curves": {},
        "points": {},
    }


def test_box_pml():
    with pytest.raises(ValueError):
        BoxPML(dim=1)


def test_box_pml_2D():
    model = BoxPML(dim=2)
    cyl = model.add_disk(0, 0, 0, 0.1, 0.1)
    cyl, model.box = model.fragment(cyl, model.box)
    model.add_physical(cyl, "cyl")
    model.add_physical(model.box, "box")
    model.set_size(model.box, 0.03)
    model.set_size(model.pmls, 0.1)
    model.set_size(cyl, 0.01)
    model.build()


def test_box_pml_3D():
    model = BoxPML(dim=3)
    sphere = model.add_sphere(0, 0, 0, 0.1)
    cyl, model.box = model.fragment(sphere, model.box)
    model.add_physical(sphere, "sphere")
    model.add_physical(model.box, "box")
    model.set_size(model.box, 0.06)
    model.set_size(model.pmls, 0.1)
    model.set_size(sphere, 0.04)
    model.build()


def test_spline():
    g = Geometry(dim=2)
    N = 20
    t = np.linspace(0, 2 * np.pi, N)
    r = np.random.rand(N) + 1
    points = np.array([r * np.cos(t), r * np.sin(t), np.zeros_like(t)]).T
    points[0, :] = points[-1, :]
    size = 0.2
    spl = g.add_spline(points, size)
    g.add_physical(spl, "kk")

    g.build()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .geometry import *


class Layered2D(Geometry):
    def __init__(
        self,
        period=1,
        thicknesses=OrderedDict(),
        **kwargs,
    ):
        super().__init__(
            dim=2,
            **kwargs,
        )
        self.period = period
        # assert isinstance(self.thicknesses == OrderedDict)
        self.thicknesses = thicknesses

        self.layers = list(thicknesses.keys())

        self.total_thickness = sum(self.thicknesses.values())

        self.y0 = -sum(list(self.thicknesses.values())[:2])
        self.layers = {}
        self.y_position = {}
        y0 = self.y0
        self._phys_groups = []
        for id, thickness in self.thicknesses.items():
            layer = self.make_layer(y0, thickness)
            self.layers[id] = layer
            self.y_position[id] = y0
            self.add_physical(layer, id)
            self._phys_groups.append(layer)
            y0 += thickness

        self.remove_all_duplicates()
        self.synchronize()

        for sub, num in self.subdomains["surfaces"].items():
            self.add_physical(num, sub)

    @property
    def translation_x(self):
        return self._translation_matrix([self.period, 0, 0])

    def make_layer(self, y_position, thickness):
        return self.add_rectangle(
            -self.period / 2, y_position, 0, self.period, thickness
        )

    def build(self, *args, **kwargs):
        s = self.get_periodic_bnds(self.y0, self.total_thickness)
        periodic_id = {k: [S[-1] for S in v] for k, v in s.items()}
        gmsh.model.mesh.setPeriodic(
            1, periodic_id["+x"], periodic_id["-x"], self.translation_x
        )
        super().build(*args, **kwargs)

    def get_periodic_bnds(self, y_position, thickness, eps=1e-3):
        pmin = -self.period / 2 - eps, -eps + y_position, -eps
        pmax = -self.period / 2 + eps, y_position + thickness + eps, eps
        s = {"-x": gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 1)}
        pmin = +self.period / 2 - eps, -eps + y_position, -eps
        pmax = +self.period / 2 + eps, y_position + thickness + eps, eps
        s["+x"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 1)

        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .geometry import *


class BoxPML2D(Geometry):
    def __init__(
        self,
        box_size=(1, 1),
        box_center=(0, 0),
        pml_width=(0.2, 0.2),
        Rcalc=0,
        **kwargs,
    ):
        super().__init__(
            dim=2,
            **kwargs,
        )
        self.box_size = box_size
        self.box_center = box_center
        self.pml_width = pml_width
        self.Rcalc = Rcalc

        def _addrect_center(rect_size):
            corner = -np.array(rect_size) / 2
            corner = tuple(corner) + (0,)
            return self.add_rectangle(*corner, *rect_size)

        def _translate(tag, t):
            translation = tuple(t) + (0,)
            self.translate(self.dimtag(tag), *translation)

        def _add_pml(s, t):
            pml = _addrect_center(s)
            _translate(pml, t)
            return pml

        box = _addrect_center(self.box_size)
        s = (self.pml_width[0], self.box_size[1])
        t = np.array([self.pml_width[0] / 2 + self.box_size[0] / 2, 0])
        pmlxp = _add_pml(s, t)
        pmlxm = _add_pml(s, -t)
        s = (self.box_size[0], self.pml_width[1])
        t = np.array([0, self.pml_width[1] / 2 + self.box_size[1] / 2])
        pmlyp = _add_pml(s, t)
        pmlym = _add_pml(s, -t)

        s = (self.pml_width[0], self.pml_width[1])
        t = np.array(
            [
                self.pml_width[0] / 2 + self.box_size[0] / 2,
                self.pml_width[1] / 2 + self.box_size[1] / 2,
            ]
        )
        pmlxypp = _add_pml(s, t)
        pmlxymm = _add_pml(s, -t)
        pmlxypm = _add_pml(s, (-t[0], t[1]))
        pmlxymp = _add_pml(s, (t[0], -t[1]))

        all_dom = [
            box,
            pmlxp,
            pmlxm,
            pmlyp,
            pmlym,
            pmlxypp,
            pmlxypm,
            pmlxymm,
            pmlxymp,
        ]
        _translate(all_dom, self.box_center)

        self.box = box
        self.pmls = all_dom[1:]

        self.fragment(self.box, self.pmls)

        self.pml_physical = [
            "pmlx",
            "pmly",
            "pmlxy",
        ]

        if Rcalc > 0:
            cyl_calc = self.add_circle(*self.box_center, 0, Rcalc)
            box, cyl_calc = self.fragment(box, cyl_calc)
            self.box = [box, cyl_calc]

        self.add_physical(box, "box")
        self.add_physical([pmlxp, pmlxm], "pmlx")
        self.add_physical([pmlyp, pmlym], "pmly")
        self.add_physical([pmlxypp, pmlxypm, pmlxymm, pmlxymp], "pmlxy")

        if Rcalc > 0:
            bnds = self.get_boundaries("box")
            self.calc_bnds = bnds[0]
            self.add_physical(self.calc_bnds, "calc_bnds", dim=1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
Sources.
"""

from abc import ABC, abstractmethod

import numpy as np
import sympy as sp
from dolfin import Constant as ConstantRe
from scipy.constants import c, epsilon_0, mu_0
from sympy.vector import CoordSys3D

from .. import dolfin
from ..complex import Complex, Constant, as_tensor, as_vector, dot, grad
from ..plot import *

_COORD = CoordSys3D("N")


def vector(components):
    return (
        components[0] * _COORD.i + components[1] * _COORD.j + components[2] * _COORD.k
    )


x = sp.symbols("x[0] x[1] x[2]", real=True)
X = vector(np.array(x))


def expression2complex_2d(expr, **kwargs):
    re, im = (p.subs(x[2], 0) for p in expr.as_real_imag())
    dexpr = [dolfin.Expression(sp.printing.ccode(p), **kwargs) for p in (re, im)]
    return Complex(*dexpr)


class Source(ABC):
    def __init__(self, wavelength, dim, phase=0, amplitude=1, degree=1, domain=None):
        self.wavelength = wavelength
        self.phase = phase
        self.amplitude = amplitude
        self.dim = dim
        self.degree = degree
        self.domain = domain

    @property
    def wavenumber(self):
        return 2 * np.pi / self.wavelength

    @property
    def pulsation(self):
        return self.wavenumber * c

    @property
    def frequency(self):
        return c / self.wavelength

    @abstractmethod
    def expression(self):
        pass

    def plot(self, figsize=None, ax=None):
        if self.dim == 2:
            tri, cb = plot(self.expression, mesh=self.domain, ax=ax)
            fig = plt.gcf()
            ax = fig.axes
            ax_ = ax[:2]
            ax_[0].set_title("Re")
            ax_[1].set_title("Im")
            for a in ax_:
                a.set_xlabel("x")
                a.set_ylabel("y")
            if figsize is not None:
                fig.set_size_inches(figsize)
            plt.tight_layout()
            return fig, ax, tri, cb
        else:
            raise NotImplementedError("plot not implemented in 3D")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


__all__ = [
    "parloop",
    "mpi_print",
]


import sys
from functools import wraps

from joblib import Parallel, delayed

from .. import dolfin


def parloop(n_jobs=1, prefer=None):
    def deco_parloop(func):
        """
        Decorate a function to parallelize.
        """

        @wraps(func)
        def my_func(*args, **kwargs):
            other_args = args[1:]
            return Parallel(n_jobs=n_jobs, prefer=prefer)(
                delayed(func)(x, *other_args, **kwargs) for x in args[0]
            )

        return my_func

    return deco_parloop


def mpi_print(*args, **kwargs):
    if dolfin.MPI.rank(dolfin.MPI.comm_world) == 0:
        print(*args, **kwargs)
        sys.stdout.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .source import *


def green_function_2d(wavelength, xs, ys, phase=0, amplitude=1, degree=1, domain=None):
    Xs = vector(sp.symbols("xs, ys, 0", real=True))
    k0 = sp.symbols("k0", real=True)
    Xshift = X - Xs
    rho = sp.sqrt(Xshift.dot(Xshift))
    rho = rho.subs(x[2], 0)
    kr = k0 * rho
    k0_ = 2 * np.pi / wavelength
    KR = dolfin.Expression(
        sp.printing.ccode(kr), k0=k0_, xs=xs, ys=ys, degree=degree, domain=domain
    )
    return (
        -1
        / 4
        * Complex(dolfin.bessel_Y(0, KR), dolfin.bessel_J(0, KR))
        * Constant(amplitude)
        * phase_shift_constant(ConstantRe(phase))
    )


class LineSource(Source):
    def __init__(
        self, wavelength, position, dim=2, phase=0, amplitude=1, degree=1, domain=None
    ):
        if dim == 3:
            raise NotImplementedError("LineSource not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.position = position

    @property
    def expression(self):
        return green_function_2d(
            self.wavelength,
            *self.position,
            phase=self.phase,
            amplitude=self.amplitude,
            degree=self.degree,
            domain=self.domain,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


def test_adjoint():
    import os

    initadj = os.environ.get("GYPTIS_ADJOINT") is not None

    import gyptis

    gyptis.use_adjoint(True)
    assert gyptis.ADJOINT is True
    assert gyptis.dolfin.__name__ == "dolfin_adjoint"

    gyptis.use_adjoint(False)
    assert gyptis.ADJOINT is False
    assert gyptis.dolfin.__name__ == "dolfin"

    if initadj:
        gyptis.use_adjoint(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
Finite element weak formulations.
"""

from abc import ABC, abstractmethod

import numpy as np
from scipy.constants import epsilon_0, mu_0

from .. import dolfin
from ..bc import *
from ..complex import *
from ..sources import *
from ..utils.helpers import project_iterative


class Formulation(ABC):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions=None,
        modal=False,
        degree=1,
        dim=1,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        self.geometry = geometry
        self.coefficients = coefficients
        self.function_space = function_space
        self.source = source
        self.trial = TrialFunction(self.function_space)
        self.test = TestFunction(self.function_space)
        self.boundary_conditions = boundary_conditions
        self.modal = modal
        self.degree = degree
        self.dim = dim  # 1: scalar problem, 3: vectorial

        self.measure = geometry.measure
        self.dx = self.measure["dx"]
        self.ds = self.measure["ds"]
        self.dS = self.measure["dS"]

        self.element = self.function_space.split()[0].ufl_element()
        self.real_function_space = dolfin.FunctionSpace(
            self.geometry.mesh, self.element
        )

    def build_lhs(self):
        self.lhs = dolfin.lhs(self.weak)
        return self.lhs

    def build_rhs(self):
        self.rhs = dolfin.rhs(self.weak)
        if self.rhs.empty():
            if self.element.value_size() == 3:
                dummy_vect = as_vector(
                    [dolfin.DOLFIN_EPS, dolfin.DOLFIN_EPS, dolfin.DOLFIN_EPS]
                )
                dummy_form = dot(dummy_vect, self.trial) * self.dx
            else:
                dummy_form = dolfin.DOLFIN_EPS * self.trial * self.dx
            self.rhs = dummy_form.real + dummy_form.imag
        return self.rhs

    def _set_rhs(self, custom_rhs):
        self.rhs = custom_rhs
        return self.rhs

    @abstractmethod
    def weak(self):
        """Weak formulation"""
        pass

    @abstractmethod
    def build_boundary_conditions(self):
        pass


def is_dolfin_function(f):
    return (
        hasattr(f.real, "ufl_shape") or hasattr(f.imag, "ufl_shape")
        if iscomplex(f)
        else hasattr(f, "ufl_shape")
    )


def find_domains_function(coeffs, list_domains=None):
    dom_function = []
    for coeff in coeffs:
        list_domains = list_domains or list(coeff.dict.keys())

        dom_function += [k for k, v in coeff.dict.items() if is_dolfin_function(v)]
    dom_function = np.unique(dom_function).tolist()
    dom_no_function = [k for k in list_domains if k not in dom_function]
    return dom_function, dom_no_function



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import pytest
from numpy import pi
from test_geometry import geom2D

from gyptis import dolfin
from gyptis.measure import *

model = geom2D(mesh_size=0.1)


def test_measure(model=model, tol=1e-9):
    dx = Measure(
        "dx",
        domain=model.mesh_object["mesh"],
        subdomain_data=model.mesh_object["markers"]["triangle"],
        subdomain_dict=model.subdomains["surfaces"],
    )

    _dx = dolfin.Measure(
        "dx",
        domain=model.mesh_object["mesh"],
        subdomain_data=model.mesh_object["markers"]["triangle"],
    )
    assert dx == _dx
    area = dolfin.assemble(1 * dx)
    assert area == dolfin.assemble(1 * _dx)
    assert abs(area - model.square_size**2) < tol
    area_cyl = dolfin.assemble(1 * dx("cyl"))
    assert abs(area_cyl - model.cyl_size**2) < tol
    area_box = dolfin.assemble(1 * dx("box"))
    assert abs(area_box - (model.square_size**2 - model.cyl_size**2)) < tol

    # exterior_facets

    ds = Measure(
        "ds",
        domain=model.mesh_object["mesh"],
        subdomain_data=model.mesh_object["markers"]["line"],
        subdomain_dict=model.subdomains["curves"],
    )

    # interior_facets

    dS = Measure(
        "dS",
        domain=model.mesh_object["mesh"],
        subdomain_data=model.mesh_object["markers"]["line"],
        subdomain_dict=model.subdomains["curves"],
    )

    len_cyl = dolfin.assemble(1 * dS("cyl_bnds"))
    assert abs(len_cyl - 4 * model.cyl_size) < tol
    len_box = dolfin.assemble(1 * ds("outer_bnds"))
    assert abs(len_box - model.square_size * 4) < tol

    area = dolfin.assemble(1 * dx("everywhere"))
    assert abs(area - model.square_size**2) < tol
    area = dolfin.assemble(1 * dx(["cyl", "box"]))
    assert abs(area - model.square_size**2) < tol

    dx = Measure(
        "dx",
        domain=model.mesh_object["mesh"],
        subdomain_data=model.mesh_object["markers"]["triangle"],
        subdomain_dict=model.subdomains["surfaces"],
        subdomain_id="cyl",
    )
    area = dolfin.assemble(1 * dx)
    assert abs(area - area_cyl) < tol



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .formulation import *


class TwoScale3D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        boundary_conditions=None,
        degree=1,
        direction="x",
        case="epsilon",
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        super().__init__(
            geometry,
            coefficients,
            function_space,
            boundary_conditions=boundary_conditions,
            degree=degree,
        )
        self.case = case
        self.epsilon, self.mu = self.coefficients
        self.direction = direction
        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)

    def poisson(self, u, v, coeff, domain="everywhere"):
        if domain == []:
            return 0
        if self.direction == "x":
            e = Constant((1, 0, 0))
        elif self.direction == "y":
            e = Constant((0, 1, 0))
        else:
            e = Constant((0, 0, 1))

        form = [inner(coeff * grad(u), grad(v))]
        form.append(dot(-coeff * e, grad(v)))
        return (form[0] + form[1]) * self.dx(domain)

    def _weak(self, u, v):
        coeff = self.epsilon if self.case == "epsilon" else self.mu
        coeff_sub = coeff.as_subdomain()
        coeff_dict = coeff.as_property()
        dom_func, dom_no_func = find_domains_function((coeff, coeff))
        form = self.poisson(u, v, coeff_sub, domain=dom_no_func)
        for dom in dom_func:
            form += self.poisson(u, v, coeff_dict[dom], domain=dom)

        return form.real + form.imag

    @property
    def weak(self):
        u = self.trial
        v = self.test
        return self._weak(u, v)

    def build_pec_boundary_conditions(self, applied_function):
        applied_function = project_iterative(applied_function, self.real_function_space)
        return build_pec_boundary_conditions(
            self.pec_boundaries,
            self.geometry,
            self.function_space,
            applied_function,
        )

    def build_boundary_conditions(self):
        applied_function = Constant(0)
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .pw import *


class GaussianBeam(Source):
    def __init__(
        self,
        wavelength,
        angle,
        waist,
        position=(0, 0),
        dim=2,
        Npw=101,
        phase=0,
        amplitude=1,
        degree=1,
        domain=None,
    ):
        if dim == 3:
            raise NotImplementedError("GaussianBeam not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.angle = angle
        self.waist = waist
        self.position = position
        self.Npw = Npw

    @property
    def expression(self):
        _expression = Constant(0)
        for t in np.linspace(-np.pi / 2, np.pi / 2, self.Npw):
            angle_i = self.angle + t
            K = self.wavenumber * np.array((-np.sin(angle_i), -np.cos(angle_i)))
            position = np.array(self.position)
            phase_pos = -np.dot(K, position)
            term = plane_wave_2d(
                self.wavelength,
                angle_i,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            ) * Constant(
                np.exp(
                    -(t**2)
                    * 4
                    * (np.pi / 2) ** 2
                    * (self.waist / self.wavelength) ** 2
                )
            )
            term *= phase_shift_constant(ConstantRe(phase_pos))
            _expression += term
        dk = np.pi / (self.Npw - 1)
        _expression *= Constant(dk)
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .source import *


def plane_wave_2d(wavelength, theta, phase=0, amplitude=1, degree=1, domain=None):
    k0 = 2 * np.pi / wavelength
    K = k0 * np.array((-np.sin(theta), -np.cos(theta)))
    K_ = vector(sp.symbols("kx, ky, 0", real=True))
    expr = amplitude * sp.exp(1j * (K_.dot(X) + phase))
    return expression2complex_2d(expr, kx=K[0], ky=K[1], degree=degree, domain=domain)


def plane_wave_3d(
    wavelength, theta, phi, psi, phase=(0, 0, 0), amplitude=1, degree=1, domain=None
):
    cx = np.cos(psi) * np.cos(theta) * np.cos(phi) - np.sin(psi) * np.sin(phi)
    cy = np.cos(psi) * np.cos(theta) * np.sin(phi) + np.sin(psi) * np.cos(phi)
    cz = -np.cos(psi) * np.sin(theta)

    k0 = 2 * np.pi / wavelength
    K = k0 * np.array(
        (
            -np.sin(theta) * np.cos(phi),
            -np.sin(theta) * np.sin(phi),
            -np.cos(theta),
        )
    )
    K_ = vector(sp.symbols("kx, ky, kz", real=True))

    phase_shifts = [np.exp((phis)) for phis in phase]
    Propp = amplitude * sp.exp(1j * (K_.dot(X)))

    code = [sp.printing.ccode(p) for p in Propp.as_real_imag()]
    prop = dolfin.Expression(
        code, kx=K[0], ky=K[1], kz=K[2], degree=degree, domain=domain
    )
    C = dolfin.as_tensor(
        [cx * phase_shifts[0], cy * phase_shifts[1], cz * phase_shifts[2]]
    )
    return Complex(prop[0] * C, prop[1] * C)


class PlaneWave(Source):
    """Short summary.

    Parameters
    ----------
    wavelength : float
        The wavelength of the plane wave.
    angle : float if `dim=2` tuple of len 3 if `dim=3`
        Angle(s) in radians.
    dim : int
        Dimension (2 or 3, the default is 2).
    phase : float
        Phase shift (in radian, the default is 0).
    amplitude : float
        Amplitude (the default is 1).
    degree : int
        Degree of the functional space (default is 1).
    domain : dolfin.cpp.mesh.Mesh
        The mesh for the domain of definition of the function.

    Attributes
    ----------
    angle

    """

    def __init__(
        self, wavelength, angle, dim=2, phase=0, amplitude=1, degree=1, domain=None
    ):
        if dim == 3 and np.isscalar(phase):
            phase = (phase, phase, phase)
        super().__init__(
            wavelength,
            dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.angle = angle

    @property
    def expression(self):
        return (
            plane_wave_2d(
                self.wavelength,
                self.angle,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
            if self.dim == 2
            else plane_wave_3d(
                self.wavelength,
                *self.angle,
                phase=self.phase,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import pytest

from gyptis.sources.stack import *


def test_stack():
    lambda0 = 1.1
    theta0 = -0.0 * pi
    phi0 = 0 * pi / 3
    psi0 = 0 * pi

    config = OrderedDict(
        {
            "superstrate": {"epsilon": 1, "mu": 1},
            "layer 1": {"epsilon": 2.3 - 0.2j, "mu": 2, "thickness": 1.7},
            "layer 2": {"epsilon": 3.5, "mu": 1, "thickness": 0.5},
            "layer 3": {"epsilon": 8.6 - 0.3j, "mu": 1, "thickness": 0.1},
            "substrate": {"epsilon": 3, "mu": 1.0},
        }
    )

    phi_, propagation_constants, efficiencies_stack = get_coeffs_stack(
        config, lambda0, theta0, phi0, psi0
    )
    print(efficiencies_stack["R"] + efficiencies_stack["T"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .ls import *


class Dipole(Source):
    def __init__(
        self,
        wavelength,
        position,
        angle=0,
        dim=2,
        phase=0,
        amplitude=1,
        degree=1,
        domain=None,
    ):
        if dim == 3:
            raise NotImplementedError("Dipole not implemented in 3D")
        super().__init__(
            wavelength,
            dim=dim,
            phase=phase,
            amplitude=amplitude,
            degree=degree,
            domain=domain,
        )
        self.position = position
        self.angle = angle

    @property
    def expression(self):
        ls = LineSource(
            self.wavelength,
            self.position,
            self.dim,
            self.phase,
            self.amplitude,
            self.degree + 1,
            self.domain,
        )
        n = as_vector(
            [ConstantRe(-np.sin(self.angle)), ConstantRe(-np.cos(self.angle))]
        )
        dls = grad(ls.expression)
        return dot(dls, n) / Constant(1j * self.wavenumber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from itertools import accumulate

import matplotlib.pyplot as plt
import numpy as np


def init_bands(sym_points, nband):
    path = []
    num_point = len(sym_points)
    for i in range(num_point - 1):
        P = sym_points[i]
        Q = sym_points[i + 1]
        _kx = np.linspace(P[0], Q[0], nband)
        _ky = np.linspace(P[1], Q[1], nband)
        k = np.vstack([_kx, _ky]).T
        if i != num_point - 2:
            k = k[:-1]
        path.append(k)
    ks = np.vstack(path)
    return ks


def init_bands_plot(sym_points, nband):
    ks = init_bands(sym_points, nband)
    dk = []
    for ik in range(len(ks) - 1):
        dk.append(np.linalg.norm(ks[ik + 1] - ks[ik]))
    ksplot = np.array(list(accumulate([0] + dk)))
    dk = []
    for ik in range(len(sym_points) - 1):
        dk.append(
            np.linalg.norm(np.array(sym_points[ik + 1]) - np.array(sym_points[ik]))
        )
    ksym = np.array(list(accumulate([0] + dk)))
    return ksplot, ksym


def plot_bands(
    sym_points,
    nband,
    eigenvalues,
    xtickslabels=None,
    color=None,
    **kwargs,
):
    if color == None:
        color = "#4d63c5"
        if "color" in kwargs:
            kwargs.pop("colors")
        if "c" in kwargs:
            kwargs.pop("c")
    ksplot, ksym = init_bands_plot(sym_points, nband)
    plt.plot(ksplot, eigenvalues, color=color, **kwargs)
    if xtickslabels is not None:
        plt.xticks(ksym, xtickslabels)
        for k in ksym:
            plt.axvline(k, c="#8a8a8a")
    plt.xlim(ksym[0], ksym[-1])
    plt.ylim(0)
    plt.ylabel(r"$\omega$")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import os

import numpy as np
import pytest

# dolfin.set_log_level(0)

ref = [1.0022840, 1.0374103, 1.0950892, 1.1977349]
r = [0.1, 0.25, 0.34, 0.43]


@pytest.mark.parametrize("i", range(4))
def test(i):
    from gyptis import dolfin
    from gyptis.models import Homogenization3D, Lattice3D

    dolfin.parameters["form_compiler"]["quadrature_degree"] = 5
    a = 1
    v = (a, 0, 0), (0, a, 0), (0, 0, a)
    R = r[i] * a
    lmin = 0.1
    eps_incl = 3

    lattice = Lattice3D(v, verbose=0)
    incl = lattice.add_sphere(a / 2, a / 2, a / 2, R)
    incl, cell = lattice.fragment(incl, lattice.cell)
    lattice.add_physical(cell, "background")
    lattice.add_physical(incl, "inclusion")
    bnd = lattice.get_boundaries(incl)
    lattice.add_physical(bnd, "bnd", 2)
    lattice.set_size("background", lmin)
    lattice.set_size("inclusion", lmin)
    lattice.set_size("bnd", lmin / 1, 2)
    # lattice.remove_all_duplicates()
    lattice.build()

    degree = 2
    epsilon = dict(background=1, inclusion=eps_incl)
    mu = dict(background=1, inclusion=1)

    hom = Homogenization3D(
        lattice,
        epsilon,
        mu,
        degree=degree,
        direct=False,
    )
    #
    eps_eff = hom.get_effective_permittivity()
    # print(eps_eff)
    n_eff = (eps_eff[0][0].real) ** 0.5
    print(ref[i])
    print(n_eff)

    assert np.allclose(n_eff, ref[i], rtol=6e-3)

    # eps = hom.formulation.epsilon.as_subdomain()
    # eps_mean = hom.unit_cell_mean(eps)
    #
    # phix = hom.solution["epsilon"]["x"]
    # n = hom.geometry.unit_normal_vector
    # psix1 = (eps_incl-1) * gy.assemble((n[0] * phix)("+")*hom.dS("bnd"))
    # eps_eff1 = eps_mean - psix1
    # print(eps_eff1.real ** 0.5)

    epsilon = dict(background=1, inclusion=1)
    mu = dict(background=1, inclusion=eps_incl)
    hom = Homogenization3D(
        lattice,
        epsilon,
        mu,
        degree=degree,
        direct=False,
    )
    #
    mu_eff = hom.get_effective_permeability()
    n_eff = (mu_eff[0][0].real) ** 0.5
    print(n_eff)
    assert np.allclose(n_eff, ref[i], rtol=6e-3)

    print("-----")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


import json
import os
from urllib.request import urlretrieve


def _get_data_path(data_path=None):
    """Return path to data dir.

    This directory stores large datasets required for the examples, to avoid
    downloading the data several times.

    By default the data dir is set to a folder named '.gyptis/data' in the
    user home folder.

    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_path : str | None
        The full path to the data dir. ``~/.gyptis/data`` by default.
    """
    if data_path is None:
        data_path = os.path.join("~", ".gyptis/data")
        data_path = os.path.expanduser(data_path)

    if not os.path.exists(data_path):
        os.makedirs(data_path)
    return data_path


def _get_config_path(config_path):
    """Return path to config file

    Parameters
    ----------
    config_path : str | None
        The path to the data dir. ``~/.gyptis`` by default.
    """
    if config_path is None:
        config_path = os.path.join("~", ".gyptis")
        config_path = os.path.expanduser(config_path)
    else:
        config_path = os.path.join(config_path, ".gyptis")
    return config_path


def _load_config(config_path):
    """Safely load a config file."""
    with open(config_path, "r") as fid:
        try:
            config = json.load(fid)
        except ValueError as e:
            # No JSON object could be decoded --> corrupt file?
            msg = f"The config file ({config_path}) is not a valid JSON file and might be corrupted"
            raise RuntimeError(msg) from e
    return config


def _set_config(config, key, value, config_file):
    """Set the configurations in the config file.

    Parameters
    ----------
    key : str
        The preference key to set.
    value : str |  None
        The value to assign to the preference key. If None, the key is
        deleted.
    config_path : str | None
        The path to the .gyptis directory.
    """
    if not isinstance(key, str):
        raise TypeError(f"key must be of type str, got {type(key)} instead")
    if not isinstance(value, str):
        raise TypeError(f"value must be of type str, got {type(value)} instead")

    if value is None:
        config.pop(key, None)
    else:
        config[key] = value

    # Write all values. This may fail if the default directory is not
    # writeable.
    config_path = os.path.dirname(config_file)
    if not os.path.isdir(config_path):
        os.mkdir(config_path)
    with open(config_file, "w") as fid:
        json.dump(config, fid, sort_keys=True, indent=0)


def download_data(
    url, data_file_name, data_key="data_dir", data_path=None, config_path=None
):
    """Downloads a remote dataset and saves path to config file.

    Checks if the data file already exists in either the path saved under the
    key ``data_key`` in the config file or the default data path;
    ``~/.gyptis/data``. If the file does not exist, downloads the data
    from ``url`` and saves to ``data_file_name`` in data_path. Finally, stores
    the location of the data in a config file, under key ``data_key``. Returns
    the path to the data file.

    Parameters
    ----------
    url : str
        Dataset URL.

    data_file_name : str
        File name to save the dataset at.


    data_path : str | None
        The path to the data dir. ``~/.gyptis/data`` by default.

    config_path: str | None
        The path to the config file. ``~/.gyptis`` by default.

    Returns
    -------
    data_file : str
        Full path of the created file.
    """

    config_path = _get_config_path(config_path)
    config_file = os.path.join(config_path, "gyptis_config.json")
    config = _load_config(config_file) if os.path.isfile(config_file) else {}
    data_path = config.get(data_key, None) or _get_data_path(data_path=data_path)
    data_file = os.path.join(data_path, data_file_name)
    # Download file if it doesn't exist
    if not os.path.exists(data_file):
        urlretrieve(url, data_file)
    # save download location in config
    _set_config(config, data_key, data_path, config_file)
    return data_file


def download_example_data(
    data_file_name, example_dir, data_key="data_dir", data_path=None, config_path=None
):
    url = (
        "https://gitlab.com/gyptis/gyptis/-/raw/master/examples/"
        + example_dir
        + "/"
        + data_file_name
    )

    return download_data(
        url,
        data_file_name,
        data_key=data_key,
        data_path=data_path,
        config_path=config_path,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import pytest

from gyptis.mesh import *


def test_marked_mesh(shared_datadir):
    mshfile = shared_datadir / "mesh.msh"
    mmsh = MarkedMesh(mshfile, 2)
    print(mmsh)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import numpy as np
import pytest


@pytest.mark.parametrize(
    "degree,polarization", [(1, "TM"), (2, "TM"), (1, "TE"), (2, "TE")]
)
def test_phc(degree, polarization):
    from gyptis import dolfin, project
    from gyptis.models.phc2d import Lattice2D, PhotonicCrystal2D

    a = 1
    v = (a, 0), (0, a)
    R = 0.2 * a
    n_eig = 6

    lattice = Lattice2D(v)
    circ = lattice.add_circle(a / 2, a / 2, 0, R)
    circ, cell = lattice.fragment(circ, lattice.cell)
    lattice.add_physical(cell, "background")
    lattice.add_physical(circ, "inclusion")

    lattice.set_size("background", 0.05)
    lattice.set_size("inclusion", 0.05)

    lattice.build()
    eps_inclusion = 8.9
    epsilon = dict(background=1, inclusion=eps_inclusion)
    mu = dict(background=1, inclusion=1)

    phc = PhotonicCrystal2D(
        lattice,
        epsilon,
        mu,
        propagation_vector=(0.1 * np.pi / a, 0.2 * np.pi / a),
        polarization=polarization,
        degree=degree,
    )
    phc.eigensolve(n_eig=n_eig, wavevector_target=0.1)
    eig_vects = phc.solution["eigenvectors"]
    mode = eig_vects[4]
    fplot = project(mode.real, phc.formulation.real_function_space)
    dolfin.plot(fplot, cmap="RdBu_r")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell2d import Maxwell2D
from .maxwell2dbands import Maxwell2DBands
from .maxwell2dperiodic import Maxwell2DPeriodic
from .maxwell3d import Maxwell3D
from .maxwell3dbands import Maxwell3DBands
from .maxwell3dperiodic import Maxwell3DPeriodic
from .twoscale2d import TwoScale2D
from .twoscale3d import TwoScale3D



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .geometry import *


class Lattice2D(Geometry):
    def __init__(
        self,
        vectors,
        **kwargs,
    ):
        super().__init__(
            dim=2,
            **kwargs,
        )
        self.vectors = vectors
        self.vertices = [
            (0, 0),
            (self.vectors[0][0], self.vectors[0][1]),
            (
                self.vectors[0][0] + self.vectors[1][0],
                self.vectors[0][1] + self.vectors[1][1],
            ),
            (self.vectors[1][0], self.vectors[1][1]),
        ]
        p = [self.add_point(*v, 0) for v in self.vertices]
        curves = [self.add_line(p[i + 1], p[i]) for i in range(3)]
        curves.append(self.add_line(p[3], p[0]))
        cl = self.add_curve_loop(curves)
        ps = self.add_plane_surface([cl])
        self.cell = ps
        # self.add_physical(self.cell, "cell")

    @property
    def translation(self):
        return (
            self._translation_matrix([*self.vectors[0], 0]),
            self._translation_matrix([*self.vectors[1], 0]),
        )

    def get_periodic_bnds(self):
        # define lines equations
        def _is_on_line(p, p1, p2):
            x, y = p
            x1, y1 = p1
            x2, y2 = p2
            if x1 == x2:
                return np.allclose(x, x1)
            else:
                return np.allclose(y - y1, (y2 - y1) / (x2 - x1) * (x - x1))

        verts = self.vertices.copy()
        verts.append(self.vertices[0])

        # get all boundaries
        bnds = self.get_entities(1)
        maps = []
        for i in range(4):
            wheres = []
            for b in bnds:
                qb = gmsh.model.getParametrizationBounds(1, b[-1])
                B = []
                for p in qb:
                    val = gmsh.model.getValue(1, b[-1], p)
                    p = val[:2]
                    belongs = _is_on_line(p, verts[i + 1], verts[i])
                    B.append(belongs)
                alls = np.all(B)
                if alls:
                    wheres.append(b)
            maps.append(wheres)
        s = {}
        s["-1"] = [m[-1] for m in maps[-1]]
        s["+1"] = [m[-1] for m in maps[1]]
        s["-2"] = [m[-1] for m in maps[0]]
        s["+2"] = [m[-1] for m in maps[2]]
        return s

    def build(self, *args, **kwargs):
        periodic_id = self.get_periodic_bnds()
        gmsh.model.mesh.setPeriodic(
            1, periodic_id["+1"], periodic_id["-1"], self.translation[0]
        )
        gmsh.model.mesh.setPeriodic(
            1, periodic_id["+2"], periodic_id["-2"], self.translation[1]
        )
        super().build(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .simulation import *


class PhotonicCrystal3D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon=None,
        mu=None,
        propagation_vector=(0, 0, 0),
        boundary_conditions=None,
        degree=1,
    ):
        if boundary_conditions is None:
            boundary_conditions = {}
        assert isinstance(geometry, Lattice3D)
        self.epsilon, self.mu = init_em_materials(geometry, epsilon, mu)
        self.periodic_bcs = Periodic3D(geometry)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "N1curl", degree, constrained_domain=self.periodic_bcs
        )
        epsilon_coeff = Coefficient(self.epsilon, geometry, degree=degree, dim=3)
        mu_coeff = Coefficient(self.mu, geometry, degree=degree, dim=3)

        coefficients = epsilon_coeff, mu_coeff
        formulation = Maxwell3DBands(
            geometry,
            coefficients,
            function_space,
            propagation_vector=propagation_vector,
            degree=degree,
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

        self.degree = degree
        self.propagation_vector = propagation_vector

    def eigensolve(self, *args, **kwargs):
        sol = super().eigensolve(*args, **kwargs)
        self.solution["eigenvectors"] = [
            u * self.formulation.phasor for u in sol["eigenvectors"]
        ]
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

from .geometry import *


class Lattice3D(Geometry):
    def __init__(
        self,
        vectors,
        periodic_tol=1e-8,
        **kwargs,
    ):
        super().__init__(
            dim=3,
            **kwargs,
        )

        self.periodic_tol = periodic_tol
        self.vectors = vectors
        v = self.vectors
        self.vertices = [
            (0, 0, 0),
            (v[0][0], v[0][1], v[0][2]),
            (
                v[0][0] + v[1][0],
                v[0][1] + v[1][1],
                v[0][2] + v[1][2],
            ),
            (v[1][0], v[1][1], v[1][2]),
            (v[2][0], v[2][1], v[2][2]),
            (v[2][0] + v[0][0], v[2][1] + v[0][1], v[2][2] + v[0][2]),
            (
                v[2][0] + v[0][0] + v[1][0],
                v[2][1] + v[0][1] + v[1][1],
                v[2][2] + v[0][2] + v[1][2],
            ),
            (v[2][0] + v[1][0], v[2][1] + v[1][1], v[2][2] + v[1][2]),
        ]
        p = [self.add_point(*v) for v in self.vertices]
        l0 = [self.add_line(p[i + 1], p[i]) for i in range(3)]
        l0.append(self.add_line(p[3], p[0]))
        cl = self.add_curve_loop(l0)
        ps0 = self.add_plane_surface([cl])

        l1 = [self.add_line(p[i + 1], p[i]) for i in range(4, 7)]
        l1.append(self.add_line(p[7], p[4]))
        cl = self.add_curve_loop(l1)
        ps1 = self.add_plane_surface([cl])
        #
        #
        l2 = [l0[0], self.add_line(p[1], p[5])]
        l2.append(l1[0])
        l2.append(self.add_line(p[4], p[0]))
        cl = self.add_curve_loop(l2)
        ps2 = self.add_plane_surface([cl])

        l3 = [l0[2], self.add_line(p[3], p[7])]
        l3.append(l1[2])
        l3.append(self.add_line(p[6], p[2]))
        cl = self.add_curve_loop(l3)
        ps3 = self.add_plane_surface([cl])

        l4 = [l2[3], l0[3], l3[1], l1[3]]
        cl = self.add_curve_loop(l4)
        ps4 = self.add_plane_surface([cl])

        l5 = [l2[1], l0[1], l3[3], l1[1]]
        cl = self.add_curve_loop(l5)
        ps5 = self.add_plane_surface([cl])
        self.perbnds = [ps0, ps1, ps2, ps3, ps4, ps5]
        sl = self.add_surface_loop(self.perbnds)
        self.cell = self.add_volume([sl])
        self.add_physical(self.cell, "cell")

        # self.cell = ps
        # self.add_physical(self.cell, "cell")

    @property
    def translation(self):
        return (
            self._translation_matrix([*self.vectors[0]]),
            self._translation_matrix([*self.vectors[1]]),
            self._translation_matrix([*self.vectors[2]]),
        )

    def get_periodic_bnds(self):
        verts = self.vertices
        # get all boundaries
        bnds = self.get_entities(2)
        maps = []
        p0 = verts[0], verts[1], verts[2]
        p1 = verts[4], verts[5], verts[6]
        p2 = verts[0], verts[1], verts[4]
        p3 = verts[2], verts[3], verts[6]
        p4 = verts[0], verts[3], verts[4]
        p5 = verts[1], verts[2], verts[6]
        self.planes = [p0, p1, p2, p3, p4, p5]
        for pl in self.planes:
            wheres = []
            for b in bnds:
                qb = gmsh.model.getParametrizationBounds(2, b[-1])
                B = []
                for p in qb:
                    P = gmsh.model.getValue(2, b[-1], p)
                    belongs = is_on_plane(P, *pl, eps=self.periodic_tol)
                    B.append(belongs)
                if alls := np.all(B):
                    wheres.append(b)
            maps.append(wheres)
        return {
            "-0": [m[-1] for m in maps[4]],
            "+0": [m[-1] for m in maps[5]],
            "-1": [m[-1] for m in maps[2]],
            "+1": [m[-1] for m in maps[3]],
            "-2": [m[-1] for m in maps[0]],
            "+2": [m[-1] for m in maps[1]],
        }

    def build(self, *args, periodic=True, **kwargs):
        if periodic:
            periodic_id = self.get_periodic_bnds()
            gmsh.model.mesh.setPeriodic(
                2, periodic_id["+0"], periodic_id["-0"], self.translation[0]
            )
            gmsh.model.mesh.setPeriodic(
                2, periodic_id["+1"], periodic_id["-1"], self.translation[1]
            )
            gmsh.model.mesh.setPeriodic(
                2, periodic_id["+2"], periodic_id["-2"], self.translation[2]
            )
        super().build(*args, **kwargs)


class CubicLattice(Lattice3D):
    def __init__(
        self,
        lattice_constant,
        **kwargs,
    ):
        self.lattice_constant = lattice_constant
        vectors = (lattice_constant, lattice_constant, lattice_constant)
        super().__init__(
            vectors,
            **kwargs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import numpy as np
import pytest


@pytest.mark.parametrize(
    "degree,polarization", [(1, "TM"), (2, "TM"), (1, "TE"), (2, "TE")]
)
def test_phc(degree, polarization):
    import gyptis
    from gyptis import BoxPML, Scattering, c, dolfin
    from gyptis.plot import plot, plot_subdomains

    wavelength = 14
    pmesh = 3
    lmin = wavelength / pmesh

    Rx, Ry = 8, 12
    eps_cyl = 41

    geom = BoxPML(
        dim=2,
        box_size=(2 * wavelength, 2 * wavelength),
        pml_width=(2 * wavelength, 2 * wavelength),
    )
    cyl = geom.add_ellipse(0, 0, 0, Rx, Ry)
    cyl, box = geom.fragment(cyl, geom.box)
    geom.add_physical(box, "box")
    geom.add_physical(cyl, "cyl")
    [geom.set_size(pml, lmin) for pml in geom.pmls]
    geom.set_size("box", lmin)
    geom.set_size("cyl", lmin / eps_cyl**0.5)
    geom.build()

    epsilon = dict(box=1, cyl=eps_cyl)
    mu = dict(box=1, cyl=1)
    s = Scattering(
        geom,
        epsilon,
        mu,
        modal=True,
        polarization=polarization,
        degree=degree,
    )
    wavelength_target = 40
    n_eig = 6
    k_target = 2 * np.pi / wavelength_target
    solution = s.eigensolve(n_eig, k_target)
    solution["eigenvalues"]
    eig_vects = solution["eigenvectors"]
    mode = eig_vects[0]
    plot(mode.real, cmap="RdBu_r", proj_space=s.formulation.real_function_space)
    H = s.formulation.get_dual(mode, 1)
    Vvect = gyptis.dolfin.VectorFunctionSpace(geom.mesh, "CG", 2)
    H = gyptis.project(
        H,
        Vvect,
        solver_type="cg",
        preconditioner_type="jacobi",
    )
    gyptis.dolfin.plot(H.real, cmap="Greys")
    geom.plot_subdomains()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io

import numpy as np
from test_geometry import geom2D

from gyptis import dolfin
from gyptis.bc import *
from gyptis.bc import _DirichletBC
from gyptis.complex import *


def test_dirichlet():
    model = geom2D(mesh_size=0.1)
    dx = model.measure["dx"]
    ds = model.measure["ds"]
    dS = model.measure["dS"]
    W = dolfin.FunctionSpace(model.mesh_object["mesh"], "CG", 1)
    boundaries_dict = model.subdomains["curves"]
    markers_line = model.mesh_object["markers"]["line"]

    ubnd = dolfin.project(dolfin.Expression("x[0]*x[1]", degree=2), W)

    bc1 = _DirichletBC(W, 1, markers_line, "outer_bnds", boundaries_dict)
    bc2 = _DirichletBC(W, ubnd, markers_line, "cyl_bnds", boundaries_dict)
    u = dolfin.TrialFunction(W)
    v = dolfin.TestFunction(W)
    a = dolfin.inner(dolfin.grad(u), dolfin.grad(v)) * dx + u * v * dx
    L = dolfin.Constant(0) * v * dx
    u = dolfin.Function(W)
    dolfin.solve(a == L, u, [bc1, bc2])
    a = dolfin.assemble(u * ds("outer_bnds"))
    assert np.abs(a - 4 * model.square_size) ** 2 < 1e-10

    dolfin.assemble(u * dS)

    r = model.cyl_size
    T = np.linspace(-r, r, 50) / 2

    U = [u(t, -r / 2) for t in T]
    exact = -r / 2 * T
    assert np.all(np.abs(np.array(U) - exact) ** 2 < 1e-6)

    U = [u(t, r / 2) for t in T]
    exact = r / 2 * T
    assert np.all(np.abs(np.array(U) - exact) ** 2 < 1e-6)

    U = [u(-r / 2, t) for t in T]
    exact = -r / 2 * T
    assert np.all(np.abs(np.array(U) - exact) ** 2 < 1e-6)

    U = [u(r / 2, t) for t in T]
    exact = r / 2 * T
    assert np.all(np.abs(np.array(U) - exact) ** 2 < 1e-6)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# Version: 1.0.2
# License: MIT
# See the documentation at gyptis.gitlab.io


import pytest
from numpy import e, pi
from test_geometry import geom2D

import gyptis
from gyptis.api import BoxPML
from gyptis.complex import *
from gyptis.geometry import *
from gyptis.materials import *

model = geom2D(mesh_size=0.1)
mesh = model.mesh_object["mesh"]
dx = model.measure["dx"]
r = model.cyl_size
lsq = model.square_size
markers = model.mesh_object["markers"]["triangle"]
domains = model.subdomains["surfaces"]


W = dolfin.FunctionSpace(mesh, "CG", 1)


@pytest.mark.parametrize("degree", [0, 1, 2])
def test_subdomain(degree):
    tol = 1e-2
    values = dict(cyl=12, box=1)
    sub = Subdomain(markers, domains, values, degree=degree)
    sub_py = Subdomain(markers, domains, values, degree=degree, cpp=False)

    a = dolfin.assemble(sub * dx)
    a_py = dolfin.assemble(sub_py * dx)
    assert a == a_py

    a_cyl = r**2
    a_box = lsq**2 - a_cyl
    a_test = a_cyl * values["cyl"] + a_box * values["box"]

    assert abs(a - a_test) ** 2 < tol

    f = dolfin.Expression("(x[0]*x[0]+x[1]*x[1])", degree=degree, r=r)

    values = dict(cyl=f, box=1)
    sub_with_function = Subdomain(markers, domains, values, degree=degree)
    integral = dolfin.assemble(sub_with_function * dx("cyl"))
    Iexact = 4 / 3 * r**3
    assert abs(integral - Iexact) ** 2 < tol
    sub_with_function_python = Subdomain(
        markers, domains, values, degree=degree, cpp=False
    )
    I_python = dolfin.assemble(sub_with_function_python * dx("cyl"))
    assert abs(I_python - Iexact) ** 2 < tol


@pytest.mark.parametrize("degree", [0, 1, 2])
def test_subdomain_complex(degree):
    tol = 1e-15
    values = dict(cyl=12 + 2j, box=1)
    sub = Subdomain(markers, domains, values, degree=degree)
    sub_py = Subdomain(
        markers,
        domains,
        values,
        degree=degree,
        cpp=False,
    )

    a = assemble(sub * dx)
    a_py = assemble(sub_py * dx)
    assert a == a_py

    a_cyl = r**2
    a_box = lsq**2 - a_cyl
    a_test = a_cyl * values["cyl"] + a_box * values["box"]
    tol1 = 1e-2
    assert abs(a - a_test) ** 2 < tol1
    f = dolfin.Expression(" exp(-pow(x[0]/r,2) - pow(x[1]/r,2))", degree=degree, r=r)

    eps_cyl = 2
    eps_box = f
    mapping = dict(cyl=eps_cyl, box=eps_box)
    eps = Subdomain(markers, domains, mapping, degree=degree)
    assert assemble((eps - eps_box) * dx("box")) < tol
    assert assemble((eps - eps_cyl) * dx("cyl")) < tol

    eps_cyl = 2 - 1.2 * 1j
    eps_box = Complex(f, f)
    mapping = dict(cyl=eps_cyl, box=eps_box)
    eps = Subdomain(markers, domains, mapping, degree=degree)
    assert assemble(abs(eps - eps_box) * dx("box")) < tol
    assert assemble(abs(eps - eps_cyl) * dx("cyl")) < tol

    eps_cyl = [[1, 2], [3, 4]]
    eps_box = 2
    mapping = dict(cyl=eps_cyl, box=eps_box)
    eps = Subdomain(markers, domains, mapping, degree=degree)
    eps_box_tens = eps_box * np.eye(2)
    for i in [0, 1]:
        for j in [0, 1]:
            assert assemble(abs(eps[i][j] - eps_box_tens[i, j]) * dx("box")) < tol
            assert assemble(abs(eps[i][j] - eps_cyl[i][j]) * dx("cyl")) < tol

    eps_cyl = [[1, 2 - 1j], [3, Complex(f, f)]]
    eps_box = 1
    mapping = dict(cyl=eps_cyl, box=eps_box)
    eps = Subdomain(markers, domains, mapping, degree=degree)
    eps_box_tens = eps_box * np.eye(2)
    for i in [0, 1]:
        for j in [0, 1]:
            assert assemble(abs(eps[i][j] - eps_box_tens[i, j]) * dx("box")) < tol
            assert assemble(abs(eps[i][j] - eps_cyl[i][j]) * dx("cyl")) < tol


def test_pml():
    PML()


def test_coefficient():
    geom = BoxPML(dim=2)
    cyl = geom.add_circle(0, 0, 0, 0.2)
    cyl, box = geom.fragment(cyl, geom.box)
    geom.add_physical(box, "box")
    geom.add_physical(cyl, "cyl")
    geom.set_size("cyl", 0.04)
    geom.build()

    epsilon = dict(box=1, cyl=3)
    eps = Coefficient(epsilon)

    pmlx = PML("x", stretch=1 - 1j, matched_domain="box", applied_domain="pmlx")
    pmly = PML("y", stretch=1 - 1j, matched_domain="box", applied_domain="pmly")
    pmlxy = PML("xy", stretch=1 - 1j, matched_domain="box", applied_domain="pmlxy")

    eps = Coefficient(epsilon, geometry=geom, pmls=[pmlx, pmly, pmlxy])
    eps.apply_pmls()

    eps.as_subdomain()

    # eps.plot()
    eps.plot(component=(1, 1))
    annex = eps.build_annex(domains=["cyl"], reference="box")

    # eps.plot()
    annex.plot(component=(1, 1))

    eps = Coefficient(epsilon, geometry=geom, pmls=[pmlx, pmly, pmlxy])

    eps.as_property()
    eps.as_property(dim=3)

    eps.to_xi()
    eps.to_chi()


# TODO: be careful here
# Warning
# /ufl/exproperators.py:336: FutureWarning: elementwise comparison failed;
# returning scalar instead, but in the future will perform elementwise comparison
# if arg in ("+", "-"):
# possible solution: check f.ufl_operands and project first
